"""
Tests for `biloba.util`.
"""

import unittest

import gevent

from biloba import util


class CachedPropertyTestCase(unittest.TestCase):
    """
    Tests for `util.cached_property`
    """

    class MyClass(object):
        @util.cachedproperty
        def foo(self):
            """
            Return a new object with a new `id` each time.
            """

            return object()

    def test_instance(self):
        """
        Simple test case for sanity
        """
        my_object = self.MyClass()

        foo = my_object.foo

        self.assertIs(foo, my_object.foo)

    def test_class(self):
        """
        Access `foo` as a class
        """
        self.assertIsInstance(
            self.MyClass.foo,
            util.cachedproperty
        )


class WaitAnyTestCase(unittest.TestCase):
    """
    Tests for `util.waitany`
    """

    def never_return(self):
        """
        A function that will never return
        """
        from gevent import event

        ev = event.Event()

        ev.wait()

    def return_sync(self):
        """
        Return a result immediately.
        """
        return 5

    def return_async(self):
        """
        Force a context switch and when the greenlet is re-activated, return
        a result.
        """
        gevent.sleep(0.0)

        return 4

    def return_error(self):
        """
        Raise an exception synchronously
        """
        raise RuntimeError('foo bar')

    def return_error_async(self):
        """
        Force a context switch and when the greenlet is re-activated, raise an
        exception
        """
        gevent.sleep(0.0)

        raise RuntimeError('foo bar')

    def test_sync(self):
        """
        Ensure that synchronous greenlets are supported.
        """
        sync = gevent.spawn(self.return_sync)

        # the context switch forces the above greenlet to run, prepping its
        # value
        gevent.sleep(0.0)

        self.assertTrue(sync.ready())

        async = gevent.spawn(self.return_async)

        ret = util.waitany([async, sync])

        self.assertIs(ret, sync)

        self.assertEqual(ret.value, 5)

    def test_async(self):
        """
        Ensure that asynchronous greenlets are supported.
        """
        async = gevent.spawn(self.return_async)
        never = gevent.spawn(self.never_return)

        ret = util.waitany([async, never])

        self.assertIs(ret, async)

        self.assertEqual(ret.value, 4)

    def test_sync_error(self):
        sync = gevent.spawn(self.return_error)
        async = gevent.spawn(self.return_async)

        ret = util.waitany([async, sync])

        self.assertIs(ret, sync)

    def test_async_error(self):
        async = gevent.spawn(self.return_error_async)
        never = gevent.spawn(self.never_return)

        ret = util.waitany([never, async])

        self.assertIs(ret, async)



"""
Provides a simple interface around dict based config objects
"""

missing = object()


class Config(object):
    """
    A wrapper around a dict to provide a simpler interface for getting
    config values. Example::

        my_config = {
            'http': {
                'address': '127.0.0.1',
                'port': 4000,
            },
            'logger': {
                'address': '${http.address}',
            }
        }

        conf = Config(my_config)

        conf.get('logger.address') == '127.0.0.1'
    """

    def __init__(self, config=None):
        self.config = config or {}

    def expand(self, value):
        if isinstance(value, list):
            return [self.expand(sub_value) for sub_value in value]

        if isinstance(value, dict):
            ret = {}

            for key, sub_value in value.items():
                ret[key] = self.expand(sub_value)

            return ret

        if not isinstance(value, basestring):
            return value

        if not value.startswith('${'):
            return value

        key = value[2:-1]

        return self.get(key)

    def get(self, key, default=None):
        value = get_key(self.config, key, default=default)

        return self.expand(value)

    def setdefault(self, key, value):
        original_value = self.get(key, missing)

        if original_value is missing:
            self.config[key] = value

    def __getitem__(self, key):
        return self.get(key)


def get_key(config, key, default=None):
    """
    Uses a dotted notation to traverse a dict.
    """
    val = config

    for part in key.split('.'):
        val = val.get(part, missing)

        if val is missing:
            return default

    return val or default


def parse_address(address, port=None):
    """
    Return a (host, port) tuple based on the config value.
    """
    try:
        address, port = address.split(':', 1)
    except ValueError:
        pass

    if port:
        port = int(port)

    return address, port
